/*
 * PCI attach
 *
 * (c) 2013, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <hw/pci.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <sys/rsrcdbmgr.h>
#include <sys/rsrcdbmsg.h>


int attach_to_device( unsigned vendor, unsigned device, int read_mem, int read_config, int stay_attached, int scan )
{
    void                *hdl;
    struct pci_dev_info pci_info;
    unsigned            *pci_mem;
    uintptr_t           pci_io;
    uint32_t            pci_config[256];
    int                 i,
                        j;

    memset( &pci_info, 0, sizeof( pci_info ) );

    if ( scan )
    {
        pci_info.BusNumber = vendor;
        pci_info.DevFunc   = device;
        hdl = pci_attach_device( NULL, PCI_SEARCH_BUSDEV | PCI_SHARE, 0, &pci_info );
    } else {
        pci_info.VendorId = vendor;
        pci_info.DeviceId = device;
        hdl = pci_attach_device( NULL, PCI_SEARCH_VENDEV | PCI_SHARE, 0, &pci_info );
    }

    if ( hdl == NULL )
    {
        if ( scan == 0 )
            perror( "Device not found:" );
        return (-1);
    }

    printf( "PCI device: %4x:%4x  Bus %4d Devfunc %#4x Class %#8x\n",
            pci_info.VendorId, pci_info.DeviceId, pci_info.BusNumber, pci_info.DevFunc, pci_info.Class );

    if ( read_config )
    {
        pci_read_config32( pci_info.BusNumber, pci_info.DevFunc, 0, 256, &pci_config );

        printf( "Dumping PCI configuration before pci_attach_device:\n" );

        for ( i = 1; i <= (256 / 4); i++ )
        {
            printf( "[0x%2X] 0x%08X ", (i - 1) * 4, pci_config[i - 1] );
            if ( i % 5 == 0 )
                printf( "\n" );
        }
        printf( "\n\n" );
    }

    if ( scan )
        return (0);
    else
        hdl = pci_attach_device( hdl, PCI_INIT_ALL | PCI_INIT_ROM | PCI_SHARE |
                                      PCI_MASTER_ENABLE | PCI_USE_MSI, 0, &pci_info );

    if ( hdl == NULL )
    {
        perror( "PCI attach failed:" );
        return (-1);
    }

    if ( read_config )
    {
        pci_read_config32( pci_info.BusNumber, pci_info.DevFunc, 0, 256, &pci_config );

        printf( "Dumping PCI configuration after pci_attach_device:\n" );
        for ( i = 1; i <= (256 / 4); i++ )
        {
            printf( "[0x%2X] 0x%08X ", (i - 1) * 4, pci_config[i - 1] );
            if ( i % 5 == 0 )
                printf( "\n" );
        }
        printf( "\n\n" );
    }

    if ( pci_info.Irq != 0xff )
        printf ("Interrupt %#x (%d)\n", pci_info.Irq, pci_info.Irq );

    for ( i = 0; i < 6; i++ )
    {
        if ( pci_info.BaseAddressSize[i] )
        {
            if ( PCI_IS_IO( pci_info.CpuBaseAddress[i] ) )
            {
                printf( "[%d] IO space at %#llx for %#x bytes\n", i,
                        PCI_IO_ADDR( pci_info.CpuBaseAddress[i] ), pci_info.BaseAddressSize[i] );

                if ( read_mem )
                {
                    pci_io = mmap_device_io( 4, PCI_IO_ADDR( pci_info.CpuBaseAddress[i] ) );

                    if ( pci_io == MAP_DEVICE_FAILED )
                    {
                        printf( "mmap_device_io for physical address %#llx failed\n",
                                PCI_IO_ADDR( pci_info.CpuBaseAddress[i] ) );
                        continue;
                    }

                    munmap_device_io( pci_io, 4 );
                }
            } else
                if ( PCI_IS_MEM( pci_info.CpuBaseAddress[i] ) )
                {
                    printf( "[%d] Memory space at %#llx for %#x bytes\n", i,
                            PCI_MEM_ADDR( pci_info.CpuBaseAddress[i] ), pci_info.BaseAddressSize[i] );

                    if ( read_mem )
                    {
                        pci_mem = mmap_device_memory( NULL, pci_info.BaseAddressSize[i],
                                                      PROT_READ | PROT_WRITE | PROT_NOCACHE,
                                                      0, PCI_MEM_ADDR( pci_info.CpuBaseAddress[i] ) );

                        if ( pci_mem == MAP_FAILED )
                        {
                            printf( "mmap_device_memory for physical address %#llx failed\n",
                                    PCI_MEM_ADDR( pci_info.CpuBaseAddress[i] ) );
                            continue;
                        }

                        printf( "Reading from device PCI memory \n" );
                        for ( j = 1; j <= 32; j++ )
                        {
                            printf( "[0x%2X] 0x%08X ", (j - 1) * 4, *(pci_mem + (j - 1)) );
                            if ( j % 5 == 0 )
                                printf( "\n" );
                        }
                        printf( "\n\n" );

                        munmap_device_memory( pci_mem, pci_info.BaseAddressSize[i] );
                    }
                }
        }
    }

    if ( pci_info.PciRom != 0x0 )
    {
        printf( "PCI ROM  0x%llx \n", PCI_ROM_ADDR( pci_info.PciRom ) );

        if ( read_mem )
        {
            pci_mem = mmap_device_memory( NULL, pci_info.RomSize,
                                          PROT_READ | PROT_WRITE | PROT_NOCACHE,
                                          0, PCI_ROM_ADDR( pci_info.PciRom ) );

            if ( pci_mem == MAP_FAILED )
            {
                printf( "mmap_device_memory for physical address %#llx failed\n",
                        PCI_ROM_ADDR( pci_info.PciRom ) );
                pci_detach_device( hdl );
                return (-1);
            }

            printf( "Reading from device PCI ROM memory : 0x%X \n", *pci_mem );
            munmap_device_memory( pci_mem, pci_info.RomSize );
        }
    }

    if ( stay_attached == 0 )
        pci_detach_device( hdl );
    else
        while ( 1 )
            sleep(1);

    return 0;
}


void dump_rsrcdbmgr( char *name, int type )
{
    rsrc_alloc_t    rsrc[256];
    int             i,
                    cnt;

    cnt = rsrcdbmgr_query( rsrc, sizeof( rsrc ) / sizeof( rsrc[0] ), 0, type );
    printf( "%s - (%d items):\n", name, cnt );
    if ( !cnt )
        return;

    printf( "%-4s %-10s %-10s %s", "Idx", "Start", "End", "Flags\n" );
    for ( i = 0; i < cnt && i < sizeof( rsrc ) / sizeof( rsrc[0] ); i++ )
        printf( "%3d: %010llx %010llx 0x%x\n",
                i, rsrc[i].start, rsrc[i].end, rsrc[i].flags );
    if ( cnt > sizeof( rsrc ) / sizeof( rsrc[0] ) )
        printf( "...\n" );
}


void dump_rsrcdbmgr_db()
{
    printf( "\nSystem resource manager database\n" );
    dump_rsrcdbmgr( "Memory",   RSRCDBMGR_MEMORY );
    dump_rsrcdbmgr( "IRQs",     RSRCDBMGR_IRQ );
    dump_rsrcdbmgr( "PCI Mem",  RSRCDBMGR_PCI_MEMORY );
    dump_rsrcdbmgr( "IO Ports", RSRCDBMGR_IO_PORT );

    printf( "\n-------------------------------------------------\n\n" );
    printf( "Used resources\n" );
    dump_rsrcdbmgr( "Memory",   RSRCDBMGR_MEMORY|RSRCDBMGR_FLAG_USED );
    dump_rsrcdbmgr( "IRQs",     RSRCDBMGR_IRQ|RSRCDBMGR_FLAG_USED );
    dump_rsrcdbmgr( "PCI Mem",  RSRCDBMGR_PCI_MEMORY|RSRCDBMGR_FLAG_USED );
    dump_rsrcdbmgr( "IO Ports", RSRCDBMGR_IO_PORT|RSRCDBMGR_FLAG_USED );
    printf( "\n" );
}


int main( int argc, char *argv[] )
{
    unsigned    vendor        = 0,
                device        = 0;
    int         read_mem      = 0,
                read_config   = 0,
                scan_flag     = 0,
                stay_attached = 0,
                system_res    = 0,
                opt,
                i,
                j,
                pd;

    if ( ThreadCtl( _NTO_TCTL_IO, 0 ) == -1 )
    {
        printf( "You must be root!" );
        return (-1);
    }

    while ( (opt = getopt( argc, argv, "sv:d:arDS" )) != -1 )
    {
        switch ( opt )
        {
            case 's':
                scan_flag = 1;
                break;

            case 'v':
                vendor = strtoul( optarg, NULL, 0 );
                break;

            case 'd':
                device = strtoul( optarg, NULL, 0 );
                break;

            case 'a':
                stay_attached = 1;
                break;

            case 'r':
                read_mem = 1;
                break;

            case 'D':
                read_config = 1;
                break;

            case 'S':
                system_res = 1;
                break;

            default:
                printf( "Unsupported option.\n" );
                return (-1);
        }
    }

    if ( system_res )
    {
        dump_rsrcdbmgr_db();
        return (0);
    }

    pd = pci_attach( 0 );
    if ( pd == (-1) )
    {
        perror( "pci_attach" );
        return (-1);
    }

    if ( scan_flag )
    {
        printf( "Scanning PCI bus\n" );
        for ( i = 0; i < 256; i++ )
            for ( j = 0; j < 256; j++ )
            {
                attach_to_device( i, j, read_mem, read_config, 0, scan_flag );
            }
    } else {
        if ( vendor == 0 || device == 0 )
            printf( "Provide correct PCI vendor and device ID.\n" );
        else {
            printf( "Attaching to PCI device 0x%04X:0x%04X \n", vendor, device );
            attach_to_device( vendor, device, read_mem, read_config, stay_attached, scan_flag );
        }
    }

    pci_detach( pd );

    return (0);
}
