/*
 * Serial port test
 *
 * (c) 2013, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <termios.h>
#ifdef __KPDANTO__
#include <sys/neutrino.h>
#endif
#include "crc16.h"


int     verbose    = 0;
char    ser_name[] = "/dev/ser1";


int write_cycle( int packet_size, int write_delay, unsigned long packet_count )
{
    int             ser_fd;
    char            *write_buffer;
    int             crc                 = 0,
                    bytes_written       = 0,
                    i,
                    k;
    unsigned long   packets_transmitted = 0;

    printf( "sertest: Starting write cycle.\n" );

    ser_fd = open( ser_name, O_WRONLY );

    if ( ser_fd == (-1) )
    {
        perror( "sertest: error opening serial port device" );
        return (-1);
    }

    write_buffer = (char *)malloc( packet_size );
    if ( write_buffer == NULL )
    {
        printf( "sertest: Unable to malloc memory for write buffer.\n" );
        return (-1);
    }

    write_buffer[0] = '[';
    k = 'a';
    for ( i = 1; i < packet_size - 1; i++ )
    {
        write_buffer[i] = k;
        if ( ++k > 'z' )
            k = 'a';
    }
    write_buffer[packet_size - 1] = ']';

    if ( verbose > 2 )
        printf( "sertest: Write buffer: \n%s\n", write_buffer );

    crc = crc16_ccitt( write_buffer + 1, packet_size - 2 );
    printf( "sertest: CRC = %4X\n", crc );

    tcflush( ser_fd, TCOFLUSH );

    while ( 1 )
    {
        bytes_written = write( ser_fd, write_buffer, packet_size );

        if ( bytes_written == (-1) )
        {
            printf( "sertest: write failed \n" );
            perror( "sertest" );
            free( write_buffer );
            return (-1);
        }

        packets_transmitted++;

        if ( verbose )
        {
            printf( "sertest: %8lu packet transmitted. %d bytes written. CRC = %4X\r",
                    packets_transmitted, bytes_written, crc );
            fflush( stdout );
        }

        if ( write_delay != 0 )
#ifdef __KPDANTO__
            delay( write_delay );
#else
            usleep( write_delay * 1000 );
#endif

        if ( packet_count != 0 && packets_transmitted == packet_count )
            break;
    }

    printf( "\nsertest: %lu packets transmitted.\n", packets_transmitted );

    close( ser_fd );
    free( write_buffer );

    return (0);
}


int read_cycle( unsigned packet_size )
{
    int             ser_fd,
                    bytes_read,
                    crc                  = 0,
                    calculated_crc       = 0,
                    received_packet_size = 0,
                    i,
                    k;
    unsigned long   received_packets     = 0,
                    bad_packets          = 0,
                    pkt_sync             = 0,
                    pkt_err              = 0;
    char            *read_buffer;
    char            *packet_start        = NULL,
                    *packet_end          = NULL;

    printf( "sertest: Starting read cycle.\n" );

    ser_fd = open( ser_name, O_RDONLY );

    if ( ser_fd == (-1) )
    {
        perror( "sertest: error opening serial port device" );
        return (-1);
    }

    read_buffer = (char *)malloc( packet_size );

    if ( read_buffer == NULL )
    {
        printf( "sertest: Unable to malloc memory for read buffer.\n" );
        return (-1);
    }

    k = 'a';
    for ( i = 1; i < packet_size - 1; i++ )
    {
        read_buffer[i] = k;
        if ( ++k > 'z' )
            k = 'a';
    }

    if ( verbose > 3 )
        printf( "sertest: \nOriginal packet \n%s\n\n", read_buffer + 1 );

    crc = crc16_ccitt( read_buffer + 1, packet_size - 2 );
    printf( "sertest: For %d bytes packet CRC = %4X \n\n", packet_size, crc );

    tcflush( ser_fd, TCIFLUSH );

    while ( 1 )
    {
        memset( read_buffer, 0, packet_size );

#ifdef __KPDANTO__
        bytes_read = readcond( ser_fd, read_buffer, packet_size, packet_size, 0, 0 );
#else
        bytes_read = read( ser_fd, read_buffer, packet_size );
#endif

        if ( bytes_read == (-1) )
        {
            printf( "sertest: read failed \n" );
            perror( "sertest" );
            free( read_buffer );
            return (-1);
        }

        if ( verbose > 1 )
        {
            printf( "sertest: \n Read buffer: \n %s\n", read_buffer );
            printf( "sertest:  %d bytes readed. \n", bytes_read );
        }

        packet_start = strchr( read_buffer, '[' );
        if ( packet_start == NULL )
        {
            printf( "sertest: Pkt start not found, skip!\n" );
            pkt_err++;
            continue;
        } else
            if ( packet_start != read_buffer )
            {
                int     t = packet_start - read_buffer;
                memcpy( read_buffer, packet_start, packet_size - t );
#ifdef __KPDANTO__
                bytes_read = readcond( ser_fd, read_buffer + packet_size - t, t, t, 0, 0 );
#else
                bytes_read = read( ser_fd, read_buffer + packet_size - t, t );
#endif
                if ( bytes_read == (-1) )
                {
                    printf( "sertest: read failed \n" );
                    perror( "sertest" );
                    free( read_buffer );
                    return (-1);
                }

                if ( verbose > 1 )
                    printf( "\nsertest: %8lu packet sync error!\n", received_packets );
                pkt_sync++;
            }

        packet_end = strchr( read_buffer, ']' );
        if ( packet_end == NULL )
        {
            printf( "\nsertest: %8lu packet end not found, skip!\n", received_packets );
            pkt_err++;
            continue;
        } else
            if ( packet_end != read_buffer + packet_size -1 )
            {
                printf( "\nsertest: %8lu packet end too short, skip!\n", received_packets );
                pkt_err++;
                continue;
            }

        if ( verbose > 3 )
            printf( "sertest: \nRead buffer: \n%s\n", read_buffer );

        calculated_crc       = 0;
        received_packet_size = packet_end - packet_start - 1;
        received_packets++;

        calculated_crc = crc16_ccitt( packet_start + 1, received_packet_size );

        if ( calculated_crc != crc )
        {
            bad_packets++;
            printf( "\nsertest: %8lu packet \t\t\t\tCRC Error!!! Calculated CRC = %4X.\n",
                    received_packets, calculated_crc );
        } else {
            if ( verbose )
                printf( "sertest: %8lu packets received Calculated CRC = %4X.\n", received_packets, calculated_crc );
            else
                printf( "sertest: %8lu packets received. %lu CRC errors \r", received_packets, bad_packets );
        }

        if ( verbose )
        {
            printf( "sertest: \t\t%4lu bad packets received. %4.2f procents of bad packets.\n",
                    bad_packets, ((float)bad_packets / (float)received_packets)*100 );

            printf( "sertest: \t\t%4lu errors, %4lu sync packets\n\n", pkt_err, pkt_sync );
        }

        fflush( stdout );
    }

    close( ser_fd );
    free( read_buffer );

    return (0);
}


int main( int argc, char *argv[] )
{
    int             opt;
    unsigned        packet_size  = 128;
    unsigned long   packet_count = 0;
    int             read_flag    = 0,
                    write_flag   = 0,
                    write_delay  = 100;

    printf( "sertest: Serial port test.\n" );

    while ( (opt = getopt( argc, argv, "p:c:wrd:vm:") ) != -1 )
    {
        switch ( opt )
        {
            case 'p':
                packet_size = strtoul( optarg, NULL, 0 );
                break;

            case 'c':
                packet_count = strtoul( optarg, NULL, 0 );
                break;

            case 'w':
                write_flag = 1;
                break;

            case 'r':
                read_flag = 1;
                break;

            case 'd':
                write_delay = strtol( optarg, NULL, 0 );
                if ( write_delay < 0 )
                    write_delay = 100;
                break;

            case 'v':
                verbose++;
                break;

            case 'm':
                strcpy( ser_name, optarg );
                break;

            default:
                return (-1);
        }
    }

    printf( "sertest: serial port device %s \n", ser_name );
    printf( "sertest: packet size = %d\n", packet_size );
    if ( packet_count )
        printf( "sertest: packet count = %lu\n", packet_count );

    if ( write_flag == 1 )
        return write_cycle( packet_size, write_delay, packet_count );
    else
        if ( read_flag == 1 )
            return read_cycle( packet_size );

    return (-1);
}
