/*
 * SPI test
 *
 * (c) 2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include "spim_test.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/resmgr.h>
#include <sys/neutrino.h>
#include <hw/inout.h>
#include <hw/nicinfo.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <inttypes.h>
#include <string.h>
#include <errno.h>
#include <devctl.h>
#include <hw/spi-master.h>


int tspi_open_cmdread( int dev, void *cmd_buf, int16_t cmd_len, void *rx_buffer, int rx_len, int *bytes_readed )
{
    int         fd;

    printf( "spi_cmdread() test...\n" );

    /* Open */
    if ( (fd = spi_open( "/dev/spi0" )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    *bytes_readed = spi_cmdread( fd, dev, cmd_buf, cmd_len, rx_buffer, rx_len );

    spi_close( fd );

    return (0);
}


int tspi_setcfg( int dev_index, uint32_t	cs, uint32_t mode, uint32_t	clock_rate )
{
    int             fd,
                    i;
    spi_cfg_t       spicfg;
    spi_devinfo_t   info;
    char            devname[20];

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    spicfg.clock_rate = clock_rate;
    spicfg.mode       = mode;

    printf( "Set configuration for SPI Device %s: cs%d  mode %x clock_rate %d \n", devname, cs, mode, clock_rate );

    if ( (fd = spi_open( devname )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    printf( "Reading existing configuration of SPI Device...\n" );

    i = spi_getdevinfo( fd, cs | SPI_DEV_DEFAULT, &info );
    if ( i == EOK )
    {
        printf( "device     = 0x%x\n", info.device );
        printf( "name       = %s\n", info.name );
        printf( "mode       = 0x%x\n", info.cfg.mode );
        printf( "clock_rate = %d\n", info.cfg.clock_rate );
    } else {
        printf( "spi_setcfg() failed!\n" );
        spi_close( fd );
        return (-1);
    }

    printf( "Changing configuration of SPI Device...\n" );

    if ( spi_setcfg( fd, cs | SPI_DEV_DEFAULT, &spicfg ) != EOK )
    {
        printf( "spi_setcfg() failed, errno = %d  %s\n", errno, strerror( errno ) );
        spi_close( fd );
        return (-1);
    }

    printf( "Reading updated configuration of SPI Device...\n" );

    i = spi_getdevinfo( fd, cs | SPI_DEV_DEFAULT, &info );
    if ( i == EOK )
    {
        printf( "device     = 0x%x\n", info.device );
        printf( "name       = %s\n", info.name );
        printf( "mode       = 0x%x\n", info.cfg.mode );
        printf( "clock_rate = %d\n", info.cfg.clock_rate );
    } else {
        printf( "spi_setcfg() failed!\n" );
        spi_close( fd );
        return (-1);
    }

    if ( (info.cfg.mode != spicfg.mode) || (info.cfg.clock_rate != spicfg.clock_rate) )
        printf( "Test failed!\n" );
    else
        printf( "Test OK!\n" );

    spi_close( fd );

    return (0);
}


int tspi_devinfo( int dev_index )
{
    spi_devinfo_t   info;
    int             i,
                    fd,
                    dev,
                    cnt;
    char            devname[20];

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    printf( "=== SPI Device %s info: ===\n", devname );

    // Open
    if ( (fd = spi_open( devname )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    printf( "Reading info using SPI_DEV_DEFAULT bit...\n" );

    dev = 0;
    cnt = 0;
    while ( 1 )
    {
        printf( "--- device 0x%04x info ---\n", dev );
        i = spi_getdevinfo( fd, dev | SPI_DEV_DEFAULT, &info );
        if ( i == EOK )
        {
            printf( "device     = 0x%x\n", info.device );
            printf( "name       = %s\n", info.name );
            printf( "mode       = 0x%x\n", info.cfg.mode );
            printf( "clock_rate = %d\n", info.cfg.clock_rate );
            dev = info.device & SPI_DEV_ID_MASK;
            cnt++;
        } else {
            printf( "Done, %d devices detected\n", cnt );
            break;
        }
        dev++;
    }

    printf( "Reading info using SPI_DEV_ID_NONE bit...\n" );

    dev = SPI_DEV_ID_NONE;
    cnt = 0;
    while ( 1 )
    {
        i = spi_getdevinfo( fd, dev, &info );
        if ( i == EOK )
        {
            printf( "device     = 0x%x\n", info.device );
            printf( "name       = %s\n", info.name );
            printf( "mode       = 0x%x\n", info.cfg.mode );
            printf( "clock_rate = %d\n\n", info.cfg.clock_rate );
            dev = info.device & SPI_DEV_ID_MASK;
            cnt++;
        } else {
            printf( "Done, %d devices detected\n", cnt );
            break;
        }
    }
    spi_close( fd );

    return (0);
}


int tspi_drvinfo( int dev_index )
{
    spi_drvinfo_t   info;
    int             i,
                    fd;
    char            devname[20];

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    printf( "=== SPI Driver %s info: ===\n", devname );

    if ( (fd = spi_open( devname )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    i = spi_getdrvinfo( fd, &info );
    if ( i == EOK )
    {
        printf( "version     = 0x%x\n", info.version );
        printf( "driver name = %s\n", info.name );
        printf( "feature     = 0x%x\n", info.feature );
    } else
        printf( "get driver info failed\n" );

    spi_close( fd );

    return (0);
}


int tspi_io_test( int dev_index, uint32_t cs, uint32_t xfer_len )
{
    int         i,
                fd,
                idx;
    char        devname[20];
    uint8_t     myBuffer[1000];
    uint8_t     rxBuffer[1000];

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    if ( xfer_len > 1000 )
        xfer_len = 1000;
    else
        if ( xfer_len == 0 )
            xfer_len = 1;

    printf( "SPI Device %s cs%d transfer test\n", devname, cs );

    if ( (fd = spi_open( devname )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    memset( myBuffer, 0xFF, 1000 );

    // Generic SPI flash read ID command
    myBuffer[0] = 0x9F;

    printf( "Transfering %d bytes of data\n", xfer_len );

    i = spi_xchange( fd, cs | SPI_DEV_DEFAULT,  (uint8_t*)myBuffer, (uint8_t*)rxBuffer, xfer_len );
    if ( i == -1 )
    {
        printf( "SPI transfer failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        spi_close( fd );
        printf( "Test failed!\n" );
        return (-1);
    }

    spi_close( fd );

    printf( "SPI transfer completed, %d byte(s) read\n", i );

    for ( idx = 0; idx < xfer_len; idx++ )
    {
        printf( "Sent data:" );
        for ( idx = 0; idx < xfer_len; idx++ )
        {
            if ( (idx & 0x1F) == 0 )
                printf( "\n" );
            printf( "%02x ", myBuffer[idx] );
        }
        printf( "\n" );

        printf( "Read data:" );
        for ( idx = 0; idx < xfer_len; idx++ )
        {
            if ( (idx & 0x1F) == 0 )
                printf( "\n" );
            printf( "%02x ", rxBuffer[idx] );
        }
        printf( "\n" );
    }

    printf( "Test completed \n" );

    return (0);
}


int tspi_iodata_test( int dev_index, uint32_t cs, uint32_t xfer_len, void *buffer )
{
    int         i,
                fd,
                idx;
    char        devname[20];
//    uint8_t     rxBuffer[1000];
//    uint8_t*    pPtr;

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    printf( "SPI Device %s cs%d transfer test\n", devname, cs );

    if ( (fd = spi_open( devname )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d  %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    printf( "Transfering %d bytes of data\n", xfer_len );

    printf( "Sent data:" );
    for ( idx = 0; idx < xfer_len; idx++ )
    {
        if ( (idx & 0x1F) == 0 )
            printf( "\n" );
        printf( "%02x ", ((uint8_t* )buffer)[idx] );
    }
    printf( "\n" );

    i = spi_xchange( fd, cs | SPI_DEV_DEFAULT,  buffer, buffer, xfer_len );
    if ( i == -1 )
    {
        printf( "SPI transfer failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        spi_close( fd );
        printf( "Test failed!\n" );
        return (-1);
    }

    spi_close( fd );

    printf( "SPI transfer completed, %d byte(s) read\n", i );

    printf( "Read data:" );
    for ( idx = 0; idx < xfer_len; idx++ )
    {
        if ( (idx & 0x1F) == 0 )
            printf( "\n" );
        printf( "%02x ", ((uint8_t *)buffer)[idx] );
    }
    printf( "\n" );

    printf( "Test completed OK!\n" );

    return (0);
}


int tspi_std_rw( int dev_index, uint32_t cs, uint32_t xfer_len )
{
    int             i,
                    fd,
                    rsize,idx;
    spi_cfg_t       spicfg;
    spi_devinfo_t   info;
//    uint8_t         b;
    char            devname[20];
    uint8_t         myBuffer[1000];
    uint8_t         rxBuffer[1000];
    uint8_t*        pPtr;

    snprintf( devname, 20, "/dev/spi%d", dev_index );

    if ( xfer_len > 1000 )
        xfer_len = 1000;
    else
        if ( xfer_len == 0 )
            xfer_len = 1;

    printf( "SPI Device %s cs%d 'std' transfer test\n", devname, cs );

    if ( (fd = open( devname, O_RDWR )) < 0 )
    {
        printf( "SPI open failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        return (-1);
    }

    printf( "Read SPI Device %s cs%d info\n", devname, cs );

    i = spi_getdevinfo( fd, cs | SPI_DEV_DEFAULT, &info );
    if ( i == -1 )
    {
        printf( "Read info failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        spi_close( fd );
        printf( "Test failed!\n" );
        return (-1);
    }

    printf( "Changing configuration of SPI Device...\n" );

    spicfg.clock_rate = info.cfg.clock_rate;
    spicfg.mode       = info.cfg.mode;

    if ( spi_setcfg( fd, cs | SPI_DEV_DEFAULT, &spicfg ) != EOK )
    {
        printf( "spi_setcfg() failed, errno = %d  %s\n", errno, strerror( errno ) );
        spi_close( fd );
        return (-1);
    }

    pPtr = (uint8_t*)myBuffer;
    for ( i = 0; i < 1000; i++ )
        *pPtr++ = i;

    printf( "Transfering %d bytes of data\n", xfer_len );

    rsize = write( fd, myBuffer, xfer_len );
    if ( rsize < 0 )
    {
        printf( "Write failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        spi_close( fd );
        printf( "Test failed!\n" );
        return (-1);
    }

    rsize = read( fd, rxBuffer, xfer_len );
    if ( rsize < 0 )
    {
        printf( "Read failed, fd = %d errno = %d %s\n", fd, errno, strerror( errno ) );
        spi_close( fd );
        printf( "Test failed!\n" );
        return (-1);
    }

    close( fd );

    printf( "SPI transfer completed, %d byte(s) read\n", rsize );

/*
    b = 0;
    for ( idx = 0; idx < xfer_len; idx++ )
        b |= rxBuffer[idx];

    if ( b == 0 )
        printf( "All readed data are 0x00. The is completed if the SPI controller not in selftest mode.\n" );
    else {
        for ( idx = 0; idx < xfer_len; idx++ )
            if ( myBuffer[idx] != rxBuffer[idx] )
            {
                printf( "Test failed!\n" );

                printf( "Sent data:\n" );
                for ( idx = 0; idx < xfer_len; idx++ )
                {
                    if ( (idx & 0x1F) == 0 )
                        printf( "\n" );
                    printf( "%02x ", myBuffer[idx] );
                }
                printf( "\n" );

                printf( "Read data:\n" );
                for ( idx = 0; idx < xfer_len; idx++ )
                {
                    if ( (idx & 0x1F) == 0 )
                        printf( "\n" );
                    printf( "%02x ", rxBuffer[idx] );
                }
                printf( "\n" );
                return (-1);
            }
    }
*/

    printf( "Sent data:" );
    for ( idx = 0; idx < xfer_len; idx++ )
    {
        if ( (idx & 0x1F) == 0 )
            printf( "\n" );
        printf( "%02x ", myBuffer[idx] );
    }
    printf( "\n" );

    printf( "Read data:" );
    for ( idx = 0; idx < xfer_len; idx++ )
    {
        if ( (idx & 0x1F) == 0 )
            printf( "\n" );
        printf( "%02x ", rxBuffer[idx] );
    }
    printf( "\n" );

    printf( "Test completed OK!\n" );

    return (0);
}


int tspi_testall( int dev_index, int cs )
{
    int         res;

    res = tspi_drvinfo( dev_index );
    if ( res == -1 )
        return res;
    res = tspi_devinfo( dev_index );
    if ( res == -1 )
        return res;

    printf( "\n--- Short frame test -- \n\n" );
    res = tspi_io_test( dev_index, cs, 10 );
    if ( res == -1 )
        return res;

    printf( "\n--- Long frame test -- \n\n" );
    res = tspi_io_test( dev_index, cs, 800 );
    if ( res == -1 )
        return res;

    printf( "\n--- 'std' test -- \n\n" );
    res = tspi_std_rw( dev_index, cs, 16 );
    if ( res == -1 )
        return res;

    printf( "\nAll tests completed OK!\n" );

    return (0);
}
