/*
 * System ticksize and timer interrupt latency control utility
 *
 * (c) 2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <inttypes.h>
#include <errno.h>
#include <sys/neutrino.h>
#include <sys/syspage.h>
#include <sys/mman.h>
#include <string.h>
#include <float.h>


#define SEC_TO_USEC         1000000
#define NSEC_TO_USEC        1000


struct sigevent         int_event;
int                     id          = -1;
int                     event_ret   = 0;
volatile unsigned long  int_counter = 0;
unsigned                dataset_size;
uint64_t                *clk_time;


const struct sigevent *handler( void *area, int id )
{
    if ( event_ret )
        return &int_event;
    else {
        clk_time[int_counter] = ClockCycles();
        int_counter++;
        if ( int_counter > dataset_size )
            return &int_event;
        else
            return (NULL);
    }
}


long get_ticksize_us()
{
    struct timespec sys_res;

    if ( clock_getres( CLOCK_REALTIME, &sys_res ) == 0 )
        return (sys_res.tv_nsec / NSEC_TO_USEC);
    else
        return (-1);
}


int set_ticksize_us( long ticksize_us )
{
    struct _clockperiod new_time;
    struct _clockperiod old_time;
    int                 ret = 0;

    memset( &new_time, 0, sizeof( struct _clockperiod ) );
    new_time.nsec = ticksize_us * NSEC_TO_USEC;
    ret = ClockPeriod( CLOCK_REALTIME, &new_time, &old_time, 0 );

    return ret;
}


void latency_test( long ticksize_us, int event_ret, unsigned time_sec, int verbose )
{
    uint64_t        cps;
    double          delta_us,
                    cps_us,
                    time_us   = 0,
                    delta_max = 0,
                    delta_avg = 0,
                    delta_min = DBL_MAX;
    unsigned int    i;

    dataset_size = time_sec * SEC_TO_USEC / ticksize_us;
    clk_time     = calloc( dataset_size, sizeof( uint64_t ) );

    cps    = SYSPAGE_ENTRY( qtime )->cycles_per_sec;
    cps_us = SEC_TO_USEC / (double)cps;

    // Attach handler to first CPU core
    ThreadCtl( _NTO_TCTL_RUNMASK, (void *)1 );

    int_event.sigev_notify = SIGEV_INTR;

    // register an interrupt handler
    id = InterruptAttach( SYSPAGE_ENTRY( qtime )->intr, &handler, NULL, 0, 0 );

    if ( id == -1 )
    {
        perror( "Cannot attach interrupt handler " );
        exit( EXIT_FAILURE );
    }

    if ( event_ret )
    {
        printf( "Measuring system timer interrupt latency for %u sec (event mode)...\n", time_sec );

        while ( int_counter < dataset_size )
        {
            InterruptWait( 0, NULL );
            clk_time[int_counter] = ClockCycles();
            int_counter++;
        }

        InterruptDetach( id );
    } else {
        printf( "Measuring system timer interrupt latency for %u sec...\n", time_sec );

        InterruptWait( 0, NULL );
        InterruptDetach( id );
    }

    for ( i = 0; i < int_counter - 1; i++ )
    {
        // Check for CPU counter overflow
        if ( clk_time[i + 1] < clk_time[i] )
            delta_us = ((double)(~(uint64_t)0 - clk_time[i] + clk_time[i + 1])) * cps_us;
        else
            delta_us = (double)(clk_time[i + 1] - clk_time[i]) * cps_us;

        if ( verbose > 1 && delta_us > (ticksize_us + 50.0) )
            printf( "idx %d delta %10.2lf stop %llx start %llx\n", i, delta_us, clk_time[i + 1], clk_time[i] );

        if ( delta_max < delta_us )
            delta_max = delta_us;

        if ( (delta_min > delta_us) && (i > 0) )
            delta_min = delta_us;

        time_us   += delta_us;
        delta_avg  = time_us / (i + 1);
    }

    if ( verbose )
    {
        printf( "\nTotal measurements: %lu \n", int_counter );
        printf( "\nMinimum = %10.2lf mks\n", delta_min );
        printf( "Maximum = %10.2lf mks\n", delta_max );
        printf( "Average = %10.2lf mks\n", delta_avg );
    }
    printf( "\n------------------------------------\n" );
    printf( "System Timer Latency = %10.2lf mks", delta_max - ticksize_us );
    printf( "\n------------------------------------\n" );
}


int main( int argc, char** argv )
{
    long        ticksize_us = 0;
    unsigned    time_sec    = 10;
    int         opt,
                verbose     = 0;

    //  Allow IO operations
    if ( ThreadCtl( _NTO_TCTL_IO, 0 ) != EOK )
    {
        printf( "You must be root!" );
        return (EXIT_FAILURE);
    }

    while ( (opt = getopt( argc, argv, "t:T:ev" )) != -1 )
    {
        switch ( opt )
        {
            case 't':
                // System timer ticksize in us
                ticksize_us = strtoul( optarg, NULL, 0 );
                break;

            case 'T':
                time_sec = strtoull( optarg, NULL, 0 );
                break;

            case 'e':
                event_ret = 1;
                break;

            case 'v':
                verbose++;
                break;

            default:
                exit( -1 );
                break;
        }
    }

    if ( ticksize_us )
    {
        if ( set_ticksize_us( ticksize_us ) != -1 )
        {
            printf( "System timer ticksize: %ld mks\n", ticksize_us );
            return (EXIT_SUCCESS);
        } else {
            printf( "Can not set system timer ticksize\n" );
            return (EXIT_FAILURE);
        }
    } else {
        if ( (ticksize_us = get_ticksize_us()) != -1 )
            printf( "System timer ticksize: %ld mks\n", ticksize_us );
        else {
            printf( "Can not get system timer ticksize\n" );
            return (EXIT_FAILURE);
        }
    }

    // Perform latency test
    latency_test( ticksize_us, event_ret, time_sec, verbose );

    return (EXIT_SUCCESS);
}
